from .nodoc import docNode
from .nodoc import docTree
from .nodoc import Node
from .nodoc import Tree
from .nodoc import vectorDB
from .nodoc import Document
from .nodoc import Markdown
from .nodoc import PDF
from .nodoc import debugger
from .nodoc import flowProcessing

__all__ = [
    'docNode',
    'docTree',
    'Node',
    'Tree',
    'vectorDB',
    'Document',
    'Markdown',
    'PDF',
    'debugger',
    'flowProcessing',
]


from typing import Any, Literal, NotRequired, TypeAlias, Required, TypedDict, Union, Self
from uuid import UUID

code_map: dict[int, str] = {
    1_200: "成功请求代码",
    1_404: "请求的内容或流程不存在",
    1_500: "流程内部错误或流程错误",
    1_503: "服务器宕机，主机强迫断开连接"
}

class MessageInit:

    class nodoc_int4(int):
        "4位整型数据，范围：[-8, 7]"

        def __init__(self, value: int) -> None:
            super().__init__()
            if not (isinstance(value, int) or (-8 < value < 7)):
                raise ValueError(f"{value} 超出范围。")

    class nodoc_int8(int):
        "8位整型数据，范围：[-128, 127]"

        def __init__(self, value: int) -> None:
            super().__init__()
            if not (isinstance(value, int) or (-128 < value < 127)):
                raise ValueError(f"{value} 超出范围。")

    class fontStylePropertyInit:

        class Color:
            ...

        class Rotation(TypedDict):
            value: float
            """
            double, 字块旋转角度 （角度值）
            """
            center: list[float]
            """
            nodoc_vec<double>[2] 二维向量, 定义字块旋转中心（相对坐标）
            """

        class Spin(TypedDict):
            value: float
            """
            double, 单个字符旋转角度 （角度值）
            """
            center: list[float]
            """
            nodoc_vec<double>[2] 二维向量，定义单个字旋转中心（相对坐标）
            """

        class Hyperlink(TypedDict):
            inherited: Union[str, "MessageInit.fontStyleProperty"]
            url: str
   


    class fontStyleProperty(TypedDict):
        color: "MessageInit.fontStylePropertyInit.Color"
        """
        字体颜色
        """
        font_size: "MessageInit.nodoc_int8"
        """
        字体尺寸
        """
        font_family: str
        """
        字体族
        """
        bold: "MessageInit.nodoc_int4"
        """
        粗体程度
        """
        underlined: bool
        """
        下划线
        """
        italic: bool
        """
        斜体
        """
        rotation: Union[float, "MessageInit.fontStylePropertyInit.Rotation"]
        """
        double | compound, 字体块旋转角度（角度值），仅是double时旋转中心默认为字块中心。
        """
        spin: Union[float, "MessageInit.fontStylePropertyInit.Spin"]
        """
        double | compound, 单个字符旋转角度（角度值），仅是double时旋转中心默认为字符中心。
        """
        hyperlink: Union[str, "MessageInit.fontStylePropertyInit.Hyperlink"]
        """
        string | compound, 该字体的超链接，当仅为字符串类型时，以url跳转到字符串中的位置。
        """
        width: float
        """
        double, bbox宽度
        """
        height: float
        """
        double, bbox高度
        """
        position: list[float]
        """
        nodoc_vec<double> ...[2] 二维向量，bbox位置（绝对坐标）
        """
        line_distance: float
        """
        double, 自动排版时正文的行间距。
        """
        font_distance: float
        """
        各个字符之间的间距。
        """
        align: Literal['left', 'right', 'center', 'justify', 'middle', 'vertical', 'horizontal', 'top', 'bottom', 'top-left', 'top-right', 'bottom-left', 'bottom-right']
        """
        string, 对齐类型
        """



    class generalProperty(TypedDict):
        inherited: Union[str, "MessageInit.fontStyleProperty"]
        """
        - Union
          - str: 继承自来自另一段指定hash消息的属性
          - fontStyleProperty: 同fontStyle的property字段一致
        """

    class titleProperty(generalProperty):
        level: "MessageInit.nodoc_int4"
        """
        unsigned nodoc_int4
        范围[0, 15]，决定标题的层级
        """

    class headerProperty(generalProperty):
        underlined: Required[bool]
        """
        不同于fontStyle的下划线，这将决定整个页眉的bbox底部是否加下划线。
        """

    class footerProperty(generalProperty):
        # Required
        overlined: Required[bool]
        """
        这将决定整个页脚的bbox顶部是否加上划线。
        """

    class textProperty(generalProperty):
        indent: float
        """
        double, 该段正文的缩进像素
        """

    class latexProperty(TypedDict):
        index: int
        """
        该段公式在父节点中的位置，用于插入公式调整排版。
        """

    class imageProperty(TypedDict):
        size: list[float]
        """
        nodoc_vec<double>[2] 二维向量，图像尺寸
        """
        position: list[float]
        """
        nodoc_vec<double>[2] 二维向量，图像位置
        """

class Checker:
    # 看到之后觉得很迷茫？
    """
    例如当type为title时，我们仅仅只有inherited和level两个property，但是由于许多type继承了generalProperty，而python无法确定类型为title时会采用哪个property字典类型，python认为以下字典中的level是不必要的，因为在其他property中没有该字段，而当我们在索引字典不存在的下标时会报错，故而python给出了警告，为了消除这个警告同时为了简化语法，故而将其封装为了一个类
    {
        "type": "title",
        "property": {
            "inherited": "...",
            "level": 3
        }
    }
    示例：
    ``` py
    import Checker
    def randomDict(x):
        if x == 0:
            return {"foo": 0}
        else:
            return {"bar": 1}
    out1 = Checker("foo", None) >> randomDict(0)
    out2 = Checker("foo", None) >> randomDict(1)
    print(out1)
    print(out2)
    ```
    输出：
    ``` py
    >> 0
       None
    ```
    """

    def __init__(self, content: str, default: Any = None) -> None:
        """
        用于解决类型提示报错。
        - content: str, 绝对存在当前type的字段
        - default: Any, 特殊情况下，该字段可能不存在值，返回某个默认值。
        """
        self.content = content
        self.default = default
        pass

    def __rshift__(self, other):
        if self.content in other:
            try:
                return other[self.content]
            except:
                return self.default
        else:
            return self.default

class MetaData(TypedDict):
    filename: str
    "文件名称"
    create: str
    "文件被创建的时间"
    access: str
    "文件最近一次被访问的时间"
    modify: str
    "文件最近一次被修改的时间"
    change: str
    "元数据最近一次被修改的时间"
    custom: dict
    "其它元数据"

class MessageUnit(TypedDict):
    type: Required[
        Literal[
            "fontStyle",
            "header",
            "footer",
            "title",
            "image",
            "latex",
            "text"
        ]
    ]
    content: NotRequired[str]
    uuid: NotRequired[UUID]
    index: NotRequired[int]
    parent: NotRequired[
        Union[
            UUID,
            list[UUID]
        ]
    ]
    property: Required[
        Union[
            MessageInit.fontStyleProperty,
            MessageInit.headerProperty,
            MessageInit.footerProperty,
            MessageInit.titleProperty,
            MessageInit.imageProperty,
            MessageInit.latexProperty,
            MessageInit.textProperty
        ]
    ]

class Message(TypedDict):
    metadata: MetaData | None
    pipeline: list[MessageUnit]
    code: int

class Data:

    def __init__(self, content: str = "") -> None:
        if not isinstance(content, str):
            raise ValueError(f"期望：str，实际：{type(content)}")
        self.__content: str = content

    @property
    def content(self) -> str:
        return self.__content

    def __lshift__(self, other: str):
        if isinstance(other, str):
            self.__content += other

        return self

    def __str__(self) -> str:
        return self.__content


document_type: TypeAlias = Literal[
    'markdown',
    'html',
    'pdf',
    'word',
    'ppt',
    'excel'
]


from .base import Document
from .markdown import Markdown
from .pdf import PDF
from ._base import Message
from ._base import Checker
from . import flow_processing as flowProcessing

__all__ = [
    'Document',
    'Markdown',
    'PDF',
    'Message',
    'Checker',
    'flowProcessing'
]



from .structure import docNode
from .structure import docTree
from .structure import Node
from .structure import Tree
from .database import vectorDB
from .document import Document
from .document import Markdown
from .document import PDF
from .document import flowProcessing

from . import debugger

__all__ = [
    'docNode',
    'docTree',
    'Node',
    'Tree',
    'vectorDB',
    'Document',
    'Markdown',
    'PDF',
    'debugger',
    'flowProcessing'
]


import time
from typing import Any, Callable, Protocol, Self, TypedDict, Union, Unpack, Literal
from .. import const

from ..document import Document

from .tree import Node
from .tree import Node, Nodes, Tree, node_return_bool
import sys


splitSign = Literal['\\']


def auto_update(func: Callable):
    def wrapper(cls: 'docTree', *args, **kwargs):
        result = func(cls, *args, **kwargs)
        cls.update()
        return result
    return wrapper


class metadata(TypedDict):
    create_time: str  # 数据的创建时间
    modify_time: str  # 数据的修改时间
    visit_time: str   # 数据的访问时间
    size: int         # 数据大小


class dataArg(TypedDict):
    head: str | None  # 数据的标头
    kind: Literal['title', 'table', 'image', 'text']
    content: str


class docArg(TypedDict):
    head: str | None    # 文档的标头
    metadata: metadata  # 元数据

class docnode_return_bool(node_return_bool):
    
    def __call__(self, node: 'docNode') -> bool:
        return super().__call__(node)


class docNode(Node):

    def __init__(self, **data: Unpack[dataArg]) -> None:
        """
        文档节点
          - **data: dataArg, 文档节点具有的属性
            - kind: Literal['title', 'table', 'image', 'text'], 节点的种类，有三种字面量，默认值为'text'
            - content: str, 节点的文本内容
        - 属性\n
            继承自Node节点
        """
        # 默认值预处理
        data.setdefault('head', None)
        data.setdefault('kind', 'text')

        super().__init__(**data)
        self.data: dataArg
        self._isTitle: bool = data['kind'] == 'title'  # 是否是标题
        self._isTable: bool = data['kind'] == 'table'  # 是否是表格
        self._isImage: bool = data['kind'] == 'image'  # 是否是图像
        self._isText: bool = data['kind'] == 'text'   # 是否是正文
        self._tree: docTree
        self._parent: docNode
        self._children: docNodes
        self._right: docNode
        self._left: docNode

    @property
    def tree(self) -> 'docTree':
        return super().tree

    @tree.setter
    def tree(self, value: 'docTree'):
        super(docTree, docTree).tree.__set__(self, value)

    @property
    def children(self) -> 'docNodes':
        return super().children

    @children.setter
    def children(self, value: 'docNodes'):
        super(docNode, docNode).children.__set__(self, value)

    @property
    def parent(self) -> 'docNode':
        return super().parent

    @parent.setter
    def parent(self, value: 'docNode'):
        super(docNode, docNode).parent.__set__(self, value)

    @property
    def left(self) -> 'docNode':
        return super().left

    @left.setter
    def left(self, value: 'docNode'):
        super(docNode, docNode).left.__set__(self, value)

    @property
    def right(self) -> 'docNode':
        return super().right

    @right.setter
    def right(self, value: 'docNode'):
        super(docNode, docNode).right.__set__(self, value)

    @property
    def isText(self) -> bool:
        "是否是正文节点"
        return self._isText

    @isText.setter
    def isText(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError(f'期望：bool，实际：{type(value)}')

        self._isText = value

    @property
    def isTitle(self) -> bool:
        "是否是标题节点"
        return self._isTitle

    @isTitle.setter
    def isTitle(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError(f'期望：bool，实际：{type(value)}')

        self._isTitle = value

    @property
    def isTable(self) -> bool:
        "是否是表格节点"
        return self._isTable

    @isTable.setter
    def isTable(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError(f'期望：bool，实际：{type(value)}')

        self._isTable = value

    @property
    def isImage(self) -> bool:
        "是否是图像节点"
        return self._isImage

    @isImage.setter
    def isImage(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError(f'期望：bool，实际：{type(value)}')

        self._isImage = value

    def __eq__(self, __value: object) -> bool:
        if __value is splitSign:
            return len(self.children) == 0
        return super().__eq__(__value)

    def __rshift__(self, other: Union['docNode', None]):
        result = super().__rshift__(other)
        if result is None:
            return None
        return result

    def __lshift__(self, other: Union['docNode', None]):
        result = super().__lshift__(other)
        if result is None:
            return None
        return result

    def get_route(self, endpoint: 'docNode', condition: docnode_return_bool | None = None) -> Union['docNodes', None]:
        """
        获取子父节点的路由 -> 默认是从该节点到目标节点之间的所有标题节点（不包括该节点）。
        - endpoint: Node，该节点的目标节点。
        """
        if condition is None:
            def __condition(node: docNode):
                    return node.isTitle
        else:
            __condition = condition
        result = super().get_route(endpoint, __condition)
        if result is None:
            return None
        return result

    def __matmul__(self, other):
        result = super().__matmul__(other)
        return result

    def __str__(self):
        return f"""
标头：{self.data['head']}
内容：{self.data['content']}
类型：{self.data['kind']}
路由：{self >> self.tree.root}
"""


class docNodes(Nodes):

    def __init__(self, nodes: 'docNodes') -> None:
        super().__init__(nodes)

    def prepend(self, node: docNode):
        return super().prepend(node)

    def append(self, node: docNode):
        return super().append(node)
    
    def remove(self, node: docNode):
        return super().remove(node)

    def __str__(self) -> str:
        return "/".join([node.data['content'] for node in self.data])
    
    def __len__(self) -> int:
        return super().__len__()


class docTree(Tree):

    def __init__(self, root: docNode, name: str = '文档树', **data: Unpack[docArg]) -> None:
        """
        实例化一个文档树对象。
        - root: docNode, 文档树的根节点。
        - name: str, 文档树的名称，用于查询。
        """
        data.setdefault('head', None)
        data.setdefault('metadata', {
            'create_time': time.strftime('%Y-%m-%d %H:%M:%S', time.localtime()),
            'modify_time': time.strftime('%Y-%m-%d %H:%M:%S', time.localtime()),
            'visit_time': time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())
        })

        if not isinstance(root, docNode):
            raise TypeError(f'期望：docNode，实际：{type(root)}')

        super().__init__(root, name)
        self.data = data
        self.root: docNode

    def update(self):
        self.data['metadata'].setdefault('size', sys.getsizeof(self.document))

    def from_document(self, document: Document):
        self.document = document.content

    # @override
    def DFT(self, node=None, callback: Callable[[Node], bool] = lambda node: True) -> list[docNode] | None:
        result = super().DFT(node, callback)
        return result

    # @override
    def BFT(self, callback: Callable[[Node], bool] = lambda node: True) -> list[docNode] | None:
        result = super().BFT(callback)
        return result

    def __str__(self):
        document = None
        if hasattr(self, 'document'):
            if len(self.document) <= 12:
                document = self.document.replace('\n', '')
            else:
                prefix = self.document[0:5].replace('\n', '')
                char_count = f'...({len(self.document) - 10}字)...'
                suffix = self.document[-6:-1].replace('\n', '')
                document = prefix + char_count + suffix

        result = f"""
{self.name}
- 创建时间：{self.data['metadata']['create_time']}
- 修改时间：{self.data['metadata']['modify_time']}
- 访问时间：{self.data['metadata']['visit_time']}
- 文档大小：{const.get_size(self.data['metadata']['size'])}
- 文档内容：{document}
"""
        return result



import os, psutil
from typing import Literal, TypeAlias, NewType, TypeVar, List

def __import():
    global ndarray, Tensor
    from numpy import ndarray
    from torch import Tensor

Embedding = TypeVar('Embedding', 'ndarray', 'Tensor', list['Tensor'])

def get_size(size: int, mode: Literal['str', 'tuple'] = 'str'):
    unit: list[str] = ['B', 'KB', 'MB', 'GB', 'TB', 'PB']
    __size = float(size)
    while __size > 1024.0 and len(unit) > 1:
        __size /= 1024.0
        unit.pop(0)
    match mode:
        case 'str':
            return f"{__size}{unit[0]}"
        case 'tuple':
            return (__size, unit[0])

def show_memory_info():
    pid = os.getpid()
    process = psutil.Process(pid)
    info = process.memory_full_info()
    memory = info.vms
    return memory

class c_source:
    directory = os.path.join(os.path.dirname(__file__), 'c_source')
    vectordb = os.path.join(directory, 'vectordb.dll')